;-----------------------------------------------------------------------------
;
; Win32 console program to display specified number of lower bits of the
; 64-bit X86 RDTSC opcode result to stdout.  paulhoule.com 3/21/2016
;
; Built with Microsoft ml.exe ver 7.1 via:
;
;   ml phrdtsc.asm /link /incremental:no /entry:main /filealign:512
;	/merge:.data=.text /merge:.rdata=.text /safeseh:no
;
; The lib environment variable must point to the folder holding kernel32.lib,
; eg  set lib=d:\msvc71\sdk\lib
;
; More recent compiler versions should work; however, their linkers may
; produce a larger executable by refusing to merge some sections.
;
;-----------------------------------------------------------------------------

	.586				;need .586 for rdtsc opcode
	.model	flat, stdcall
	includelib kernel32.lib

GetCommandLineA proto			;Win32 definitions
GetStdHandle	proto	:dword
STD_OUTPUT_HANDLE equ	-11
WriteFile	proto	hFile:ptr, :ptr char, :dword, :ptr dword, :ptr

SkipToArgs	proto			;local function prototypes
GetNumBits	proto
FmtDec64	proto

	.data
NL	textequ	<0dh,0ah>
errmsg	db	'*** bad argument',NL
helpmsg	db	NL
db 'Displays lower n (0..64) bits of CPU timer.  paulhoule.com 3/21/2016',NL
db 'eg: phrdtsc 31',NL
endmsg	db	0 dup(?)

	.code

main	proc
	local	buf[32]:byte		;scratch buffer

	invoke	GetStdHandle,STD_OUTPUT_HANDLE
	xchg	edi,eax			;edi= standard output handle
	invoke	GetCommandLineA
	xchg	esi,eax			;esi= ptr to Win32 command line

	call	SkipToArgs		;advance esi to first arg
	call	GetNumBits		;eax= parsed first arg (0..64)
	.if	eax > 64		;if invalid (or missing) argument,
	  mov	ebx,offset errmsg	;ebx= invalid arg + help msg
	  or	eax,eax
	  .if	sign?			;if no argument present
	    add ebx,sizeof errmsg	;ebx= help only msg
	  .endif
	  mov	esi,offset endmsg	;esi= length of msg
	  sub	esi,ebx

	.else
	  xchg	ecx,eax			;ecx= bits to return (0..64)
	  rdtsc
	  cmp	ecx,64			;edx:eax= isolated lower bits
	  sbb	esi,esi
	  shl	esi,cl
	  not	esi
	  .if	ecx < 32
	    and	eax,esi
	    cdq
	  .endif
	  and	edx,esi

	  lea	ebx,buf[sizeof buf-2]	;ebx= ptr to formatted result + crlf
	  mov	word ptr [ebx],0a0dh
	  call	FmtDec64
	  lea	esi,[eax+2]
	.endif

	invoke	WriteFile,edi,ebx,esi,addr buf,0
	ret
main	endp

;-----------------------------------------------------------------------------
; Skips over program name to first argument of Win32 command line.  This
; simple logic depends on the observed behavior that Win32 insures at least
; one space exists between the program name and first argument.
;
; In:	esi= ptr to ANSI command line string (from GetCommandLineA)
; Out:	esi= advanced to first argument, or terminating 0 if no args present
;-----------------------------------------------------------------------------

SkipToArgs proc
	xor	eax,eax
	.repeat
	  lodsb
	  .break .if !al
	  .if	al == '"'
	    lodsb
	    .break .if !al
	    .continue .if al == '"'	;adjacent double quotes are ignored
	    not	ah
	  .endif
	  or	al,ah			;ignore if within double quotes
	.until	al == ' '

	.while al == ' ' || al == 9	;skip all space/tab after prog name
	  lodsb
	.endw

	dec	esi
	ret
SkipToArgs endp

;-----------------------------------------------------------------------------
; Parses number of bits argument.  Stops if value exceeds max valid (64) or
; a non-digit is encountered.  Argument must end with 0/space/tab.
;
; In:	esi= ptr within 0-terminated string to start of argument
; Out:	esi= advanced beyond processed digit(s) -- this may be no advance
;	eax= 0..64 - valid result, 65 - any overflow or bad terminator,
;		eax = 80000000h if valid terminator but no digits found
;-----------------------------------------------------------------------------

GetNumBits proc
	xor	eax,eax			;set all digit register bits zero
	mov	edx,80000000h		;init result to "no digits seen"

	.repeat
	  lodsb
	  sub	al,'0'
	  .break .if carry? || al > 9
	  imul	edx,10
	  add	edx,eax
	.until	edx > 64

	add	al,'0'
	.if	!zero? && al != ' ' && al != 9
	  mov	edx,65			;overflow or bad terminator - error
	.endif

	dec	esi			;esi= ptr to terminator char
	xchg	eax,edx
	ret
GetNumBits endp

;-----------------------------------------------------------------------------
; Formats a 64-bit unsigned to decimal.
;
; In:	edx:eax= 64-bit value (treated as unsigned)
;	ebx= 1+ ptr for LOWEST ORDER digit -- ie last output buffer byte + 1
; Out:	ebx= ptr to highest order digit... at least 1 digit is always stored
;	eax= output digit count (1..20)
;-----------------------------------------------------------------------------

FmtDec64 proc	uses edi
	push	ebx
	push	10			;ecx= a ten to use as divisor
	pop	ecx

	mov	edi,edx			;loop requires dividend in edi:edx
	.repeat
	  xor	edx,edx
	  xchg	eax,edi
	  div	ecx
	  xchg	eax,edi
	  div	ecx

	  dec	ebx
	  add	dl,'0'
	  mov	[ebx],dl

	  mov	edx,edi
	  or	edx,eax
	.until	zero?

	pop	eax			;eax= number of digits output
	sub	eax,ebx
	ret
FmtDec64 endp

	end
